import re
from logging.handlers import SocketHandler

from pygelf import GelfUdpHandler, GelfTcpHandler, GelfTlsHandler


class SocketHandler(SocketHandler):
    def __init__(self, host: str, port: int | None):
        super().__init__(host, port)  # pragma: no cover

    def emit(self, record):
        """
        Emit a record.

        Pickles the record and writes it to the socket in binary format.
        If there is an error with the socket, silently drop the packet.
        If there was a problem with the socket, re-establishes the
        socket.
        """

        # parse out ansi color tags..
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        record.msg = ansi_escape.sub("", record.msg)

        try:
            s = self.makePickle(record)
            self.send(s)
        except Exception:  # pragma: no cover
            self.handleError(record)


class GelfUDPHandler(GelfUdpHandler, SocketHandler):
    def __init__(self, host, port, **kwargs):
        super().__init__(host, port, **kwargs)


class GelfTCPHandler(GelfTcpHandler, SocketHandler):
    def __init__(self, host, port, **kwargs):
        super().__init__(host, port, **kwargs)  # pragma: no cover


class GelfTLSHandler(GelfTlsHandler, SocketHandler):
    def __init__(self, host, port, **kwargs):
        super().__init__(host, port, **kwargs)  # pragma: no cover



import logging
import sys

from CveXplore.common.config import Configuration
from CveXplore.core.logging.handlers.cve_explore_rfh import CveExploreUpdateRfhHandler
from CveXplore.core.logging.handlers.cve_explore_stream import (
    CveExploreUpdateStreamHandler,
)


class UpdateBaseClass(object):
    def __init__(self, logger_name: str):
        self.config = Configuration
        self.logger = logging.getLogger(logger_name)

        if len(self.logger.handlers) == 1:
            self.logger.removeHandler(self.logger.handlers[0])

        self.logger.propagate = False

        self.formatter = logging.Formatter(
            "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
        )

        crf = None

        cli = CveExploreUpdateStreamHandler(stream=sys.stdout)
        cli.setFormatter(self.formatter)
        cli.setLevel(logging.INFO)

        if self.config.LOGGING_TO_FILE:
            crf = CveExploreUpdateRfhHandler(
                filename=f"{self.config.LOGGING_FILE_PATH}/{self.config.LOGGING_UPDATE_FILE_NAME}",
                maxBytes=self.config.LOGGING_MAX_FILE_SIZE,
                backupCount=self.config.LOGGING_BACKLOG,
            )
            crf.setLevel(logging.DEBUG)
            crf.setFormatter(self.formatter)

        if len(self.logger.handlers) > 0:
            for handler in self.logger.handlers:
                # add the handlers to the logger
                # makes sure no duplicate handlers are added
                if not isinstance(
                    handler, CveExploreUpdateRfhHandler
                ) and not isinstance(handler, CveExploreUpdateStreamHandler):
                    if crf is not None:
                        self.logger.addHandler(crf)
                    self.logger.addHandler(cli)
        else:
            if crf is not None:
                self.logger.addHandler(crf)
            self.logger.addHandler(cli)



import json
import os

from dotenv import load_dotenv

user_wd = os.path.expanduser("~/.cvexplore")

load_dotenv(os.path.join(user_wd, ".env"))

import contextlib
import logging
import time

from celery import Celery
from celery.backends.database import SessionManager
from celery.signals import (
    task_postrun,
    task_retry,
    task_failure,
    task_prerun,
    setup_logging,
)
from celery.utils.log import get_task_logger

from CveXplore.common.config import Configuration
from CveXplore.core.logging.logger_class import AppLogger
from CveXplore.core.general.constants import task_status_codes
from CveXplore.core.redis_handler.redis_handler import RedisHandler

logging.setLoggerClass(AppLogger)

config = Configuration

app = Celery(
    "CveXplore",
    broker=f"{config.CELERY_REDIS_URL}{config.CELERY_REDIS_BROKER_DB}",
    backend=f"{config.CELERY_REDIS_URL}{config.CELERY_REDIS_BACKEND_DB}",
    result_extended=True,
    include=["CveXplore.celery_app.cvexplore_daemon"],
    task_time_limit=config.CELERY_TASK_TIME_LIMIT,
    task_default_queue="default",
    broker_connection_retry_on_startup=True,
    result_expires=config.CELERY_RESULT_EXPIRES,
)

execution_times = {}


@setup_logging.connect
def setup_logging(logger, *args, **kwargs):
    # Disregard all celery processing on loggers and let the generic loggerClass handle the configuration of the loggers
    pass


@task_prerun.connect
def general_task_pre_run_config(task_id, task, *args, **kwargs):
    logger = get_task_logger(__name__)

    execution_times[task_id] = time.time()

    task.update_state(state="STARTED")

    logger.info("Task started!")


@task_retry.connect
def general_task_retry_config(request, reason, einfo, *args, **kwargs):
    logger = get_task_logger(__name__)

    logger.warning(f"Task retry initiated reason: {reason}")


@task_postrun.connect
def general_task_post_run_config(task_id, task, retval, state, *args, **kwargs):
    logger = get_task_logger(__name__)

    try:
        cost = time.time() - execution_times.pop(task_id)
    except KeyError:
        cost = -1

    logger.info(
        f"Task post run cost: {cost}, State: {state}, Task: {task}, RetVal: {retval}"
    )

    task.request.update(task_execution_time=cost)

    if "task_slug" in task.request.kwargs:
        task_slug = task.request.kwargs["task_slug"]
    else:
        task_slug = task.name

    if isinstance(retval, Exception) or retval is None:
        task.backend.client.set(
            f"runresult_{task_slug}",
            config.CELERY_TASK_FAILED_ERROR_CODE,
            ex=86400,
        )
        insert_time = int(time.time())
        task.backend.client.zadd(
            f"sortresults_{task_slug}", {f"{task_slug}_{task_id}": insert_time}
        )
        task.backend.client.hset(
            f"{task_slug}_{task_id}",
            mapping={
                "state": state,
                "cost": cost,
                "returns": json.dumps(
                    {
                        "status": config.CELERY_TASK_FAILED_ERROR_CODE,
                        "data": {"out": "", "err": f"{type(retval)}"},
                    }
                ),
                "inserted": insert_time,
                "task_id": task_id,
            },
        )
        task.backend.client.expire(
            name=f"{task_slug}_{task_id}",
            time=config.CELERY_KEEP_TASK_RESULT * 60 * 60 * 24,
        )
    else:
        task.backend.client.set(f"runresult_{task_slug}", retval["status"], ex=86400)
        insert_time = int(time.time())
        task.backend.client.zadd(
            f"sortresults_{task_slug}", {f"{task_slug}_{task_id}": insert_time}
        )
        task.backend.client.hset(
            f"{task_slug}_{task_id}",
            mapping={
                "state": state,
                "cost": cost,
                "returns": json.dumps(retval),
                "inserted": insert_time,
                "task_id": task_id,
            },
        )
        task.backend.client.expire(
            name=f"{task_slug}_{task_id}",
            time=config.CELERY_KEEP_TASK_RESULT * 60 * 60 * 24,
        )

    logger.info("Task execution completed!")


@task_failure.connect
def general_task_failure_config(task_id, exception, traceback, einfo, *args, **kwargs):
    logger = get_task_logger(__name__)

    logger.exception(exception)


@app.on_after_configure.connect
def setup_periodic_tasks(sender, **kwargs):
    # do tasks every xx seconds
    # sender.add_periodic_task(5.0, crt_test.s())
    pass


@contextlib.contextmanager
def get_db_session():
    session_logger = logging.getLogger(__name__)

    session_manager = SessionManager()
    engine, Session = session_manager.create_session(config.SQLALCHEMY_DATABASE_URI)
    session = Session()

    try:
        yield session
    except Exception as err:
        session_logger.warning(f"Error inserting data into database: {err}")
        session_logger.exception(err)
        session.rollback()
    finally:
        session.close()


@app.task(
    autoretry_for=(Exception,),
    max_retries=5,
    retry_backoff=True,
    retry_backoff_max=700,
    retry_jitter=True,
    ignore_result=True,
)
def crt_test(task_slug: str, *args, **kwargs) -> dict:
    """
    General test task with random response

    Args:
        task_slug: Task slug
        args: Task Arguments
        kwargs: Task Keyword arguments

    Returns:
        A dictionary with key "status" with will reflect the status of the task

    Group:
        tasks

    """
    import random

    logger = get_task_logger(__name__)
    logger.info("Testing with OK / NOK response")

    a = random.randrange(20)

    if a % 2 == 0:
        return {"status": task_status_codes.OK}
    else:
        return {"status": task_status_codes.NOK}


@app.task(
    ignore_result=True,
    task_time_limit=1800,
)
def crt_update(task_slug: str, *args, **kwargs) -> dict:
    """
    Task to automatically update the database. Task will only run one instance at the time to prevent the update tasks
    interfering with each other.

    Args:
        task_slug: Task slug
        args: Task Arguments
        kwargs: Task Keyword arguments

    Returns:
        A dictionary with key "status" with will reflect the status of the task

    Group:
        tasks

    """
    from CveXplore import CveXplore

    rh = RedisHandler(redis_client=app.backend.client)

    # Acquiring a lock to prevent this specific task to run twice or more at a time
    with rh.acquire_lock(f"crt_update"):
        try:
            cvex = CveXplore()
            cvex.database.update()
            return {"status": task_status_codes.OK}
        except Exception:
            return {"status": task_status_codes.NOK}



import logging
from abc import ABC, abstractmethod

from CveXplore.core.logging.logger_class import AppLogger

logging.setLoggerClass(AppLogger)


class DatabaseConnectionBase(ABC):
    def __init__(self, logger_name: str):
        self.logger = logging.getLogger(logger_name)

    def __repr__(self):
        return f"<<{self.__class__.__name__}>>"

    @property
    @abstractmethod
    def dbclient(self):
        raise NotImplementedError

    @abstractmethod
    def set_handlers_for_collections(self):
        raise NotImplementedError



import argparse
import logging
import os
import sys
from collections import namedtuple
from subprocess import run, PIPE, STDOUT, CompletedProcess

from CveXplore.core.logging.logger_class import AppLogger

logging.setLoggerClass(AppLogger)


class DatabaseMigrator(object):
    def __init__(self, cwd: str = None):
        self.logger = logging.getLogger(__name__)

        self.current_dir = (
            cwd
            if cwd is not None
            else os.path.dirname(
                os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
            )
        )

        self._commands = namedtuple(
            "commands", "INIT REVISION UPGRADE CURRENT HISTORY REV_UP REV_DOWN"
        )(1, 2, 3, 4, 5, 6, 7)

    @property
    def commands(self) -> namedtuple:
        return self._commands

    def db_init(self) -> None:
        res = self.__cli_runner(self.commands.INIT)
        self.__parse_command_output(res)

    def db_revision(self, message: str) -> None:
        res = self.__cli_runner(self.commands.REVISION, message=message)
        self.__parse_command_output(res)

    def db_upgrade(self) -> None:
        res = self.__cli_runner(self.commands.UPGRADE)
        self.__parse_command_output(res)

    def db_current(self) -> None:
        res = self.__cli_runner(self.commands.CURRENT)
        self.__parse_command_output(res)

    def db_history(self) -> None:
        res = self.__cli_runner(self.commands.HISTORY)
        self.__parse_command_output(res)

    def db_up(self, count: int) -> None:
        res = self.__cli_runner(self.commands.REV_UP, message=count)
        self.__parse_command_output(res)

    def db_down(self, count: int) -> None:
        res = self.__cli_runner(self.commands.REV_DOWN, message=count)
        self.__parse_command_output(res)

    def __parse_command_output(self, cmd_output: CompletedProcess) -> None:
        if cmd_output.returncode != 0:
            self.logger.error(cmd_output.stdout)
        else:
            output_list = cmd_output.stdout.split("\n")

            for m in output_list:
                if m != "":
                    self.logger.info(m)

    def __cli_runner(self, command: int, message: str | int = None) -> CompletedProcess:
        if command == 2 and message is None:
            raise ValueError("Missing message for revision command")
        elif command == 6 and message is None:
            raise ValueError(
                "You must specify a positive number when submitting a upgrade command"
            )
        elif command == 7 and message is None:
            raise ValueError(
                "You must specify a negative number when submitting a downgrade command"
            )

        command_mapping = {
            1: f"alembic init alembic",
            2: f'alembic revision --autogenerate -m "{message}"',
            3: f"alembic upgrade head",
            4: f"alembic current",
            5: f"alembic history --verbose",
            6: f"alembic upgrade {message}",
            7: f"alembic downgrade {message}",
        }
        try:
            result = run(
                command_mapping[command],  # nosec
                stdout=PIPE,
                stderr=STDOUT,
                universal_newlines=True,
                shell=True,
                cwd=self.current_dir,
            )
            return result
        except KeyError:  # pragma: no cover
            self.logger.error(f"Unknown command number received....")

    def __repr__(self) -> str:
        return f"<< {self.__class__.__name__} >>"


if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description="migrate/update the database schema"
    )

    argparser.add_argument(
        "-i", action="store_true", help="Setup new alembic environment"
    )
    argparser.add_argument(
        "-r", action="store_true", help="Create new revision the database"
    )
    argparser.add_argument(
        "-u", action="store_true", help="Update the database to latest head"
    )

    argparser.add_argument(
        "-up", action="store_true", help="Upgrade the database x revisions"
    )
    argparser.add_argument(
        "-down", action="store_true", help="Downgrade the database x revisions"
    )

    argparser.add_argument("-cs", action="store_true", help="Print current state")
    argparser.add_argument("-hist", action="store_true", help="Print history")

    args = argparser.parse_args()

    fsm = DatabaseMigrator()

    if (args_count := len(sys.argv)) < 2 and args.r:
        print("You must specify a message when submitting a new revision")
        raise SystemExit(2)
    elif args_count < 2 and args.up:
        print("You must specify a positive number when submitting a upgrade command")
        raise SystemExit(2)
    elif args_count < 2 and args.down:
        print("You must specify a negative number when submitting a downgrade command")
        raise SystemExit(2)

    if args.i:
        fsm.db_init()

    if args.r:
        fsm.db_revision(sys.argv[1])

    if args.u:
        fsm.db_upgrade()

    if args.up:
        fsm.db_up(int(sys.argv[1]))

    if args.down:
        fsm.db_down(int(sys.argv[1]))

    if args.cs:
        fsm.db_current()

    if args.hist:
        fsm.db_history()



import os
import shutil

from dotenv import load_dotenv

if not os.path.exists(os.path.expanduser("~/.cvexplore")):
    os.mkdir(os.path.expanduser("~/.cvexplore"))

user_wd = os.path.expanduser("~/.cvexplore")

if not os.path.exists(os.path.join(user_wd, ".env")):
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), "common/.env_example"),
        os.path.join(user_wd, ".env"),
    )

load_dotenv(os.path.join(user_wd, ".env"))

if not os.path.exists(os.path.join(user_wd, ".sources.ini")):
    shutil.copyfile(
        os.path.join(os.path.dirname(__file__), "common/.sources.ini"),
        os.path.join(user_wd, ".sources.ini"),
    )

import functools
import json
import logging

import re
import warnings
from collections import defaultdict
from typing import List, Tuple, Union, Iterable, Type

from pymongo import DESCENDING

from CveXplore.api.connection.api_db import ApiDatabaseSource
from CveXplore.common.config import Configuration
from CveXplore.common.cpe_converters import create_cpe_regex_string
from CveXplore.common.db_mapping import database_mapping
from CveXplore.core.database_maintenance.main_updater import MainUpdater
from CveXplore.core.general.datasources import supported_datasources
from CveXplore.database.connection.database_connection import DatabaseConnection
from CveXplore.errors import DatabaseIllegalCollection
from CveXplore.errors.datasource import UnsupportedDatasourceException
from CveXplore.errors.validation import CveNumberValidationError
from CveXplore.objects.cvexplore_object import CveXploreObject

try:
    from CveXplore.core.celery_task_handler.task_handler import TaskHandler
except ModuleNotFoundError:
    # This is not a problem, because using the Celery backend is optional.
    # This exception will be handled when loading the TaskHandler fails with
    # a NameError, because we already have logging available there.
    pass


def _version():
    _PKG_DIR = os.path.dirname(__file__)
    version_file = os.path.join(_PKG_DIR, "VERSION")
    with open(version_file, "r") as fdsec:
        VERSION = fdsec.read()

    return VERSION


VERSION = __version__ = _version()


class CveXplore(object):
    """
    The CveXplore class is the main entry point for CveXplore. All functionality is available from straight from this
    class or via the different subclasses / attributes.

    Group:
        Main
    """

    def __init__(self, **kwargs):
        """
        Create a new instance of CveXplore

        Examples:

            >>> from CveXplore import CveXplore
            >>> cvx = CveXplore(datasource_type="mongodb", datasource_connection_details={"host": "mongodb://127.0.0.1:27017"})
            >>> cvx.version
            '0.1.2'

            >>> from CveXplore import CveXplore
            >>> cvx = CveXplore(datasource_type="api", datasource_connection_details={"address": ("mylocal.cve-search.int", 443), "api_path": "api"})
            >>> cvx.version
            '0.1.2'

        Args:
            kwargs['datasource_type']: Which datasource to query.
            kwargs['datasource_connection_details']: Provide the connection details needed to establish a connection \
            to the datasource. The connection details should be in line with the datasource's documentation.

        """
        self.__version = VERSION
        self._config = Configuration
        self.logger = logging.getLogger(__name__)

        # adjust self.config with kwargs parameters; lower case parameters are converted to uppercase and then changed
        # in the self.config object
        for each in kwargs:
            setattr(self.config, each.upper(), kwargs[each])

        self._datasource_type = self.config.DATASOURCE_TYPE

        self._datasource_connection_details = self.config.DATASOURCE_CONNECTION_DETAILS

        self._mongodb_connection_details = self.config.MONGODB_CONNECTION_DETAILS
        self._api_connection_details = self.config.API_CONNECTION_DETAILS

        os.environ["DOC_BUILD"] = json.dumps({"DOC_BUILD": "NO"})

        self.logger.info(
            f"Using {self.datasource_type} as datasource, connection details: {self.datasource_connection_details}"
        )

        if self.datasource_type not in supported_datasources:
            raise UnsupportedDatasourceException(
                f"Unsupported datasource selected: '{self.datasource_type}'; currently supported: {supported_datasources}"
            )

        if self.datasource_type == "api" and self.datasource_connection_details is None:
            raise ValueError(
                "Missing datasource_connection_details for selected datasource ('api')"
            )
        elif self.datasource_type == "api":
            self.datasource_connection_details["user_agent"] = (
                f"CveXplore:{self.version}"
            )

        if (
            self.mongodb_connection_details is not None
            and self.datasource_type == "mongodb"
        ):
            self.logger.warning(
                "The use of mongodb_connection_details is deprecated and will be removed in the 0.4 release, please "
                "use datasource_connection_details instead"
            )
            self._datasource_connection_details = self.mongodb_connection_details
        elif self.api_connection_details is not None and self.datasource_type == "api":
            self.logger.warning(
                "The use of api_connection_details is deprecated and will be removed in the 0.4 release, please "
                "use datasource_connection_details instead"
            )
            self._datasource_connection_details = self.api_connection_details
        else:
            if self.datasource_type == "mongodb":
                self._datasource_connection_details = {
                    "host": (
                        f"{self.config.DATASOURCE_PROTOCOL}://{self.config.DATASOURCE_HOST}:{self.config.DATASOURCE_PORT}"
                        if self.config.DATASOURCE_USER is None
                        and self.config.DATASOURCE_PASSWORD is None
                        else f"{self.config.DATASOURCE_PROTOCOL}://{self.config.DATASOURCE_USER}:{self.config.DATASOURCE_PASSWORD}@{self.config.DATASOURCE_HOST}:{self.config.DATASOURCE_PORT}"
                    ),
                    "database": self.config.DATASOURCE_DBNAME,
                }
            elif self.datasource_type == "mysql":
                self._datasource_connection_details = {
                    "sql_uri": self.config.SQLALCHEMY_DATABASE_URI
                }
            elif self.datasource_type == "api":
                self._datasource_connection_details = {
                    "address": (
                        f"{self.config.DATASOURCE_HOST}",
                        int(f"{self.config.DATASOURCE_PORT}"),
                    ),
                    "api_path": "api",
                }

        setattr(
            self.config,
            "DATASOURCE_CONNECTION_DETAILS",
            self.datasource_connection_details,
        )

        self.datasource = DatabaseConnection(
            database_type=self.datasource_type,
            database_init_parameters=self.datasource_connection_details,
        ).database_connection
        if self.datasource_type != "api":
            self.database = MainUpdater(
                datasource=self.datasource, datasource_type=self.datasource_type
            )

        self._database_mapping = database_mapping

        from CveXplore.database.helpers.specific_db import (
            CvesDatabaseFunctions,
            CpeDatabaseFunctions,
            CapecDatabaseFunctions,
            CWEDatabaseFunctions,
        )

        self.cves = CvesDatabaseFunctions(collection="cves")
        self.cpe = CpeDatabaseFunctions(collection="cpe")
        self.capec = CapecDatabaseFunctions(collection="capec")
        self.cwe = CWEDatabaseFunctions(collection="cwe")

        try:
            self.task_handler = TaskHandler()
        except NameError:
            self.logger.info(
                "Failed to load Celery TaskHandler; "
                "continuing without the optional Celery backend."
            )

        self.logger.info(f"Initialized CveXplore version: {self.version}")

    @property
    def config(self) -> Type[Configuration]:
        """
        Returns Configuration object

        Group:
            Properties
        """
        return self._config

    @property
    def datasource_type(self) -> str:
        """
        Returns Datasource type

        Group:
            Properties
        """
        return self._datasource_type

    @property
    def datasource_connection_details(self) -> dict:
        """
        Returns Datasource Connection Details

        Group:
            Properties
        """
        return self._datasource_connection_details

    @property
    def database_mapping(self) -> dict:
        """
        Returns database mapping

        Group:
            Properties
        """
        return self._database_mapping

    @property
    def mongodb_connection_details(self) -> dict:
        """
        Returns Mongodb Connection Details

        Group:
            Properties
        """
        warnings.warn(
            "The use of mongodb_connection_details is deprecated and will be removed in the 0.4 release, "
            "please use datasource_connection_details instead",
            DeprecationWarning,
        )
        return self._mongodb_connection_details

    @property
    def api_connection_details(self) -> dict:
        """
        Returns Api Connection Details

        Group:
            Properties
        """
        warnings.warn(
            "The use of api_connection_details is deprecated and will be removed in the 0.4 release, please "
            "use datasource_connection_details instead",
            DeprecationWarning,
        )
        return self._api_connection_details

    @property
    def version(self) -> str:
        """
        Returns CveXplore version

        Group:
            Properties
        """
        return self.__version

    @staticmethod
    def where() -> str:
        """
        Request the path where CveXplore is installed

        Returns:
            Path where CveXplore is installed
        """
        return os.path.dirname(__file__)

    def get_single_store_entry(
        self, entry_type: str, dict_filter: dict = None
    ) -> CveXploreObject | None:
        """
        Method to perform a query on a *single* collection in the data source and return a *single* result.

        Examples:

            >>> from CveXplore import CveXplore
            >>> cvx = CveXplore()
            >>> result = cvx.get_single_store_entry("capec", {"id": "1"})
            >>> result
            << Capec:1 >>

        Args:
            entry_type: Which specific store are you querying? Choices are:

                                - capec;
                                - cpe;
                                - cwe;
                                - via4;
                                - cves;
            dict_filter: a dictionary representing a filter according to pymongo documentation

        Returns:
            An instance of CveXploreObject or None
        """

        if dict_filter is None:
            dict_filter = {}

        entry_type = entry_type.lower()

        if entry_type not in self.database_mapping:
            raise DatabaseIllegalCollection(
                f"Illegal collection requested: only {self.database_mapping} are allowed!"
            )

        result = getattr(self.datasource, f"store_{entry_type}").find_one(dict_filter)

        return result

    def get_single_store_entries(
        self, query: Tuple[str, dict], limit: int = 10
    ) -> List[CveXploreObject] | None:
        """
        Method to perform a query on a *single* collection in the data source and return all the results.

        Examples:

           >>> from CveXplore import CveXplore
           >>> cvx = CveXplore()
           >>> result = cvx.get_single_store_entries(("cves", {"cvss": {"$eq": 8}}))
           >>> result
           [<< Cves:CVE-2011-0387 >>,
           << Cves:CVE-2015-1935 >>,
           << Cves:CVE-2014-3053 >>,
           << Cves:CVE-2010-4031 >>,
           << Cves:CVE-2016-1338 >>,
           << Cves:CVE-2013-3633 >>,
           << Cves:CVE-2017-14444 >>,
           << Cves:CVE-2017-14446 >>,
           << Cves:CVE-2017-14445 >>,
           << Cves:CVE-2016-2354 >>]

           >>> from CveXplore import CveXplore
           >>> cvx = CveXplore()
           >>> result = cvx.get_single_store_entries(("cves", {"cvss": {"$eq": 8}}), limit=0)
           >>> len(result)
           32

        Args:
            query: Tuple which contains the entry_type and the dict_filter in a tuple. Choices for entry_type:

                        - capec;
                        - cpe;
                        - cwe;
                        - via4;
                        - cves;

                    dict_filter is a dictionary representing a filter according to pymongo documentation.
            limit: Limit the number of results to return.

        Returns:
            A list with CveXploreObject's or None
        """
        if not isinstance(query, tuple):
            raise ValueError(
                f"Wrong parameter type, received: {type(query)} expected: tuple"
            )

        if len(query) != 2:
            raise ValueError(
                f"Query parameter does not consist of the expected amount of variables, "
                f"expected: 2 received: {len(query)}"
            )

        entry_type, dict_filter = query

        entry_type = entry_type.lower()

        if entry_type not in self.database_mapping:
            raise DatabaseIllegalCollection(
                f"Illegal collection requested: only {self.database_mapping} are allowed!"
            )

        if entry_type == "cves":
            if "id" in dict_filter:
                if isinstance(dict_filter["id"], str):
                    dict_filter["id"] = self._validate_cve_id(dict_filter["id"])
                elif isinstance(dict_filter["id"], dict):
                    if "$in" in dict_filter["id"]:
                        dict_filter["id"]["$in"] = [
                            self._validate_cve_id(x) for x in dict_filter["id"]["$in"]
                        ]
            if "cvss" in dict_filter:
                if isinstance(dict_filter["cvss"], str):
                    dict_filter["cvss"] = float(dict_filter["cvss"])
            if "cvss3" in dict_filter:
                if isinstance(dict_filter["cvss3"], str):
                    dict_filter["cvss3"] = float(dict_filter["cvss3"])
            if "cvss4" in dict_filter:
                if isinstance(dict_filter["cvss4"], str):
                    dict_filter["cvss4"] = float(dict_filter["cvss4"])

            if "exploitabilityScore" in dict_filter:
                if isinstance(dict_filter["exploitabilityScore"], str):
                    dict_filter["exploitabilityScore"] = float(
                        dict_filter["exploitabilityScore"]
                    )
            if "exploitabilityScore3" in dict_filter:
                if isinstance(dict_filter["exploitabilityScore3"], str):
                    dict_filter["exploitabilityScore3"] = float(
                        dict_filter["exploitabilityScore3"]
                    )

            if "impactScore" in dict_filter:
                if isinstance(dict_filter["impactScore"], str):
                    dict_filter["impactScore"] = float(dict_filter["impactScore"])
            if "impactScore3" in dict_filter:
                if isinstance(dict_filter["impactScore3"], str):
                    dict_filter["impactScore3"] = float(dict_filter["impactScore3"])

            if "epss" in dict_filter:
                if isinstance(dict_filter["epss"], str):
                    dict_filter["epss"] = float(dict_filter["epss"])

        results = (
            getattr(self.datasource, f"store_{entry_type}")
            .find(dict_filter)
            .limit(limit)
        )

        the_results = list(results)

        if len(the_results) != 0:
            return the_results
        else:
            return None

    def get_multi_store_entries(
        self, *queries: List[Tuple[str, dict]], limit: int = 10
    ) -> List[CveXploreObject] | None:
        """
        Method to perform *multiple* queries on *a single* or *multiple* collections in the data source and return the
        results.

        Examples:

            >>> from CveXplore import CveXplore
            >>> cvx = CveXplore()
            >>> result = cvx.get_multi_store_entries([("CWE", {"id": "78"}), ("cves", {"id": "CVE-2009-0018"})])
            >>> result
            [<< Cwe:78 >>, << Cves:CVE-2009-0018 >>]

        Args:
            queries: A list of tuples which contains the entry_type and the dict_filter. Choices for entry_type:

                    - capec;
                    - cpe;
                    - cwe;
                    - via4;
                    - cves;

                    dict_filter is a dictionary representing a filter according to pymongo documentation.
            limit: Limit the number of results to return.

        Returns:
            A list with CveXploreObject's or None
        """

        results = map(
            functools.partial(self.get_single_store_entries, limit=limit), *queries
        )

        the_results = [
            result_list for result_list in results if result_list is not None
        ]

        # flatten results
        the_results = [item for row in the_results for item in row]

        if len(the_results) != 0:
            return the_results
        else:
            return None

    def cves_for_cpe(self, cpe_string: str) -> List[CveXploreObject] | None:
        """
        Method for retrieving Cves that match a single CPE string. By default, the search will be made matching
        the configuration fields of the cves documents.

        Examples:

            >>> from CveXplore import CveXplore
            >>> cvx = CveXplore()
            >>> cves = cvx.cves_for_cpe("cpe:2.3:o:microsoft:windows_7:*:sp1:*:*:*:*:*:*")

        Args:
            cpe_string: CPE string to search for.

        Returns:
            A list with CveXploreObject's or None
        """

        cpe_regex_string = create_cpe_regex_string(cpe_string)

        cves = self.get_single_store_entries(
            ("cves", {"vulnerable_configuration": {"$regex": cpe_regex_string}}),
            limit=0,
        )

        return cves

    def _validate_cve_id(self, cve_id: str) -> str:
        """
        Method for validating a CVE ID.

        Args:
            cve_id: The CVE ID to validate.

        Returns:
            A validated CVE ID.
        """
        reg_match = re.compile(r"[cC][vV][eE]-\d{4}-\d{4,10}")
        if reg_match.match(cve_id) is not None:
            cve_id = cve_id.upper()
        else:
            part_match = re.compile(r"\d{4}-\d{4,10}")
            if part_match.match(cve_id) is not None:
                cve_id = f"CVE-{cve_id}"
            else:
                raise CveNumberValidationError(
                    f"Could not validate the CVE number: {cve_id}. The number format should be either "
                    "CVE-2000-0001, cve-2000-0001 or 2000-0001."
                )

        return cve_id

    def cve_by_id(self, cve_id: str) -> CveXploreObject | None:
        """
        Method to retrieve a single CVE from the database by its CVE ID number.

        Args:
            cve_id: The number format should be either CVE-2000-0001, cve-2000-0001 or 2000-0001.

        Returns:
            A CveXploreObject or None
        """
        cve_id = self._validate_cve_id(cve_id=cve_id)

        return self.get_single_store_entry("cves", {"id": cve_id})

    def cves_by_id(self, *cve_ids: str) -> Union[Iterable[CveXploreObject], Iterable]:
        """
        Method to retrieve multiple CVE's from the database by their CVE ID numbers.

        Args:
            cve_ids: The number format should be either CVE-2000-0001, cve-2000-0001 or 2000-0001.

        Returns:
            A list with CveXploreObject's or empty list
        """
        ret_data = []
        for cve_id in cve_ids:
            ret_data.append(self.cve_by_id(cve_id=cve_id))

        if len(ret_data) >= 1:
            return sorted(ret_data, key=lambda x: x.id)
        else:
            return ret_data

    def capec_by_cwe_id(self, cwe_id: int) -> List[CveXploreObject] | None:
        """
        Method to retrieve CAPEC's related to a specific CWE ID

        Args:
            cwe_id: The CWE ID to retrieve CAPEC for.

        Returns:
            A list with CveXploreObject's or None
        """
        cwe = self.get_single_store_entry("cwe", {"id": cwe_id})

        if cwe is not None:
            return list(cwe.iter_related_capecs())
        else:
            return cwe

    def last_cves(self, limit: int = 10) -> List[CveXploreObject] | None:
        """
        Method to retrieve the last entered / changed cves.

        Args:
            limit: The amount of results to return; defaults to 10.

        Returns:
            A list with CveXploreObject's or None
        """
        results = (
            getattr(self.datasource, "store_cves")
            .find()
            .sort("modified", DESCENDING)
            .limit(limit)
        )

        the_results = list(results)

        if len(the_results) != 0:
            return the_results
        else:
            return None

    def get_db_content_stats(self) -> dict | str:
        """
        This method request the statistics from the database and consist of the time last modified and the document count
        per cvedb store in the database.

        Returns:
            The stats from the database.
        """
        stats = defaultdict(dict)

        if not isinstance(self.datasource, ApiDatabaseSource):
            if hasattr(self.datasource, "store_info"):
                results = self.datasource.store_info.find({})
                for each in results:
                    each.pop("_id")
                    db = each["db"]
                    if db == "epss":
                        continue

                    each.pop("db")

                    if "sources" in each:
                        each.pop("sources")
                        each.pop("searchables")

                    each["lastModified"] = str(each["lastModified"])
                    each["document count"] = getattr(
                        self.datasource, f"store_{db}"
                    ).count_documents({})
                    stats[db] = each

                    for mgmtlist in ["mgmt_blacklist", "mgmt_whitelist"]:
                        stats[mgmtlist] = {
                            "document count": getattr(
                                self.datasource, f"store_{mgmtlist}"
                            ).count_documents({})
                        }

                return dict(stats)

            else:
                return "Database info could not be retrieved"

        return f"Using api endpoint: {self.datasource.baseurl}"

    def __repr__(self) -> str:
        """
        String representation of object

        Returns:
            String representation of object
        """
        return f"<< CveXplore:{self.version} >>"



import ast
import json
import os
from json import JSONDecodeError


def getenv_bool(name: str, default: str = "False"):
    raw = os.getenv(name, default).title()
    try:
        the_bool = ast.literal_eval(raw)

        if not isinstance(the_bool, bool):
            raise ValueError
    except ValueError:
        raise

    return the_bool


def getenv_list(name: str, default: list = None):
    if default is None:
        default = []

    raw = os.getenv(name, default)

    if not isinstance(raw, list):
        try:
            the_list = json.loads(raw.replace("\n", ""))
            return the_list
        except JSONDecodeError:
            raise

    return default


def getenv_dict(name: str, default: dict = None):
    if default is None:
        default = {}

    raw = os.getenv(name, default)

    if not isinstance(raw, dict):
        try:
            the_dict = json.loads(raw)
            return the_dict
        except JSONDecodeError:
            raise

    return default


class Configuration(object):
    """
    Class holding the configuration
    """

    if json.loads(os.getenv("DOC_BUILD", '{"DOC_BUILD": "NO"}'))["DOC_BUILD"] == "YES":
        USER_HOME_DIR = "${HOME}/.cvexplore"
    else:
        USER_HOME_DIR = os.path.expanduser("~/.cvexplore")

    CVE_START_YEAR = int(os.getenv("CVE_START_YEAR", 2000))

    CPE_FILTER_DEPRECATED = getenv_bool("CPE_FILTER_DEPRECATED", "True")

    # Which datasource to query.Currently supported options include:
    # - mongodb
    # - api
    DATASOURCE_TYPE = os.getenv("DATASOURCE_TYPE", "mongodb")

    DATASOURCE_PROTOCOL = os.getenv("DATASOURCE_PROTOCOL", "mongodb")
    DATASOURCE_DBAPI = os.getenv("DATASOURCE_DBAPI", None)
    DATASOURCE_HOST = os.getenv(
        "DATASOURCE_HOST", os.getenv("MONGODB_HOST", "127.0.0.1")
    )
    DATASOURCE_PORT = int(
        os.getenv("DATASOURCE_PORT", int(os.getenv("MONGODB_PORT", 27017)))
    )

    DATASOURCE_USER = os.getenv("DATASOURCE_USER", None)
    DATASOURCE_PASSWORD = os.getenv("DATASOURCE_PASSWORD", None)
    DATASOURCE_DBNAME = os.getenv("DATASOURCE_DBNAME", "cvedb")

    DATASOURCE_CONNECTION_DETAILS = None

    SQLALCHEMY_DATABASE_URI = os.getenv(
        "SQLALCHEMY_DATABASE_URI",
        (
            f"{DATASOURCE_PROTOCOL}://{DATASOURCE_USER}:{DATASOURCE_PASSWORD}@{DATASOURCE_HOST}:{DATASOURCE_PORT}/{DATASOURCE_DBNAME}"
            if DATASOURCE_DBAPI is None
            else f"{DATASOURCE_PROTOCOL}+{DATASOURCE_DBAPI}://{DATASOURCE_USER}:{DATASOURCE_PASSWORD}@{DATASOURCE_HOST}:{DATASOURCE_PORT}/{DATASOURCE_DBNAME}"
        ),
    )
    SQLALCHEMY_TRACK_MODIFICATIONS = getenv_bool(
        "SQLALCHEMY_TRACK_MODIFICATIONS", "False"
    )
    SQLALCHEMY_ENGINE_OPTIONS = getenv_dict(
        "SQLALCHEMY_ENGINE_OPTIONS", {"pool_recycle": 299, "pool_timeout": 20}
    )

    # keep these for now to maintain backwards compatibility
    API_CONNECTION_DETAILS = None
    MONGODB_CONNECTION_DETAILS = None
    MONGODB_HOST = os.getenv("MONGODB_HOST", "127.0.0.1")
    MONGODB_PORT = int(os.getenv("MONGODB_PORT", 27017))

    DEFAULT_SOURCES = {
        "cwe": "https://cwe.mitre.org/data/xml/cwec_latest.xml.zip",
        "capec": "https://capec.mitre.org/data/xml/capec_latest.xml",
        "via4": "https://www.cve-search.org/feeds/via4.json",
        "epss": "https://epss.cyentia.com/epss_scores-current.csv.gz",  # See EPSS at https://www.first.org/epss
    }

    if os.getenv("SOURCES") is not None:
        SOURCES = getenv_dict("SOURCES", None)
    else:
        try:
            with open(os.path.join(USER_HOME_DIR, ".sources.ini")) as _f:
                SOURCES = json.loads(_f.read())
        except FileNotFoundError:
            SOURCES = DEFAULT_SOURCES

    NVD_NIST_API_KEY = os.getenv("NVD_NIST_API_KEY", None)
    NVD_NIST_NO_REJECTED = getenv_bool("NVD_NIST_NO_REJECTED", "True")
    NVD_NIST_TIMEOUT = float(os.getenv("NVD_NIST_TIMEOUT", 30.0))

    HTTP_PROXY_DICT = getenv_dict("HTTP_PROXY_DICT", {})
    HTTP_PROXY_STRING = os.getenv("HTTP_PROXY_STRING", "")

    LOGGING_TO_FILE = getenv_bool("LOGGING_TO_FILE", "True")
    LOGGING_FILE_PATH = os.getenv(
        "LOGGING_FILE_PATH", os.path.join(USER_HOME_DIR, "log")
    )

    try:
        if not os.path.exists(LOGGING_FILE_PATH):
            os.mkdir(LOGGING_FILE_PATH)
    except FileNotFoundError:
        # Probably creating documentation; just pass
        pass

    LOGGING_MAX_FILE_SIZE = (
        int(os.getenv("LOGGING_MAX_FILE_SIZE", 100)) * 1024 * 1024
    )  # in MB
    LOGGING_BACKLOG = int(os.getenv("LOGGING_BACKLOG", 5))
    LOGGING_FILE_NAME = os.getenv("LOGGING_FILE_NAME", "./cvexplore.log")
    LOGGING_UPDATE_FILE_NAME = os.getenv("LOGGING_FILE_NAME", "update_populate.log")
    LOGGING_LEVEL = os.getenv("LOGGING_LEVEL", "DEBUG")

    SYSLOG_ENABLE = getenv_bool("SYSLOG_ENABLE", "False")
    SYSLOG_SERVER = os.getenv("SYSLOG_SERVER", "172.16.1.1")
    SYSLOG_PORT = int(os.getenv("SYSLOG_PORT", 5140))
    SYSLOG_LEVEL = os.getenv("SYSLOG_LEVEL", "DEBUG")

    GELF_SYSLOG = getenv_bool("GELF_SYSLOG", "True")
    # GELF format allows for additional fields to be submitted with each log record; Key values of this dict should
    # start with underscores; e.g. {"_environment": "SPECIAL"} would append an environment field with the value of
    # 'SPECIAL' to each log record.
    GELF_SYSLOG_ADDITIONAL_FIELDS = getenv_dict("GELF_SYSLOG_ADDITIONAL_FIELDS", None)

    MAX_DOWNLOAD_WORKERS = int(os.getenv("MAX_DOWNLOAD_WORKERS", 10))

    # This factor determines the amount of simultaneous requests made towards the NIST API;
    # The set amount of client requests (30) get divided with the sem factor, so the lower
    # it is set, the more simultaneous requests are made.
    DOWNLOAD_SEM_FACTOR = float(
        os.getenv("DOWNLOAD_SEM_FACTOR", 0.0)
    )  # if set, should be set >=0.6
    DOWNLOAD_SLEEP_MIN = float(os.getenv("DOWNLOAD_SLEEP_MIN", 0.5))
    DOWNLOAD_SLEEP_MAX = float(os.getenv("DOWNLOAD_SLEEP_MAX", 2.5))
    DOWNLOAD_BATCH_RANGE = os.getenv("DOWNLOAD_BATCH_RANGE", None)

    REDIS_URL = os.getenv("REDIS_URL", "redis://127.0.0.1:6379/")

    # CELERY SETTINGS
    CELERY_REDIS_URL = os.getenv("CELERY_REDIS_URL", "redis://redis:6379/")
    CELERY_REDIS_BROKER_DB = int(os.getenv("CELERY_REDIS_BROKER_DB", 5))
    CELERY_REDIS_BACKEND_DB = int(os.getenv("CELERY_REDIS_BACKEND_DB", 6))

    CELERY_TASK_FAILED_ERROR_CODE = int(
        os.getenv("CELERY_TASK_FAILED_ERROR_CODE", 1337)
    )
    CELERY_KEEP_TASK_RESULT = int(os.getenv("CELERY_KEEP_TASK_RESULT", 7))  # in days

    CELERY_TASK_TIME_LIMIT = int(os.getenv("CELERY_TASK_TIME_LIMIT", 900))  # in seconds
    CELERY_RESULT_EXPIRES = int(os.getenv("CELERY_RESULT_EXPIRES", 300))  # in seconds

    def __repr__(self):
        return f"<< CveXploreConfiguration >>"



import re

from CveXplore.database.helpers.cpe_conversion import cpe_uri_to_fs, cpe_fs_to_uri


def split_cpe_name(cpename: str) -> list[str]:
    """
    Split CPE 2.3 into its components, accounting for escaped colons.
    """
    non_escaped_colon = r"(?<!\\):"
    split_name = re.split(non_escaped_colon, cpename)
    return split_name


def from2to3CPE(cpe: str, autofill: bool = False) -> str:
    """
    Method to transform cpe2.2 to cpe2.3 format
    """
    cpe = cpe.strip()
    if not cpe.startswith("cpe:2.3:"):
        if not cpe.startswith("cpe:/"):
            return False
        cpe = cpe_uri_to_fs(cpe)
    if autofill:
        e = split_cpe_name(cpe)
        for x in range(0, 13 - len(e)):
            cpe += ":-"
    return cpe


def from3to2CPE(cpe: str) -> str:
    """
    Method to transform cpe2.3 to cpe2.2 format
    """
    cpe = cpe.strip()
    if not cpe.startswith("cpe:/"):
        if not cpe.startswith("cpe:2.3:"):
            return False
        cpe = cpe_fs_to_uri(cpe)
    return cpe


def create_cpe_regex_string(str_input: str) -> str:
    # format to cpe2.3
    cpe_string = from2to3CPE(str_input)

    if cpe_string.startswith("cpe"):
        # strict search with term starting with cpe; e.g: cpe:2.3:o:microsoft:windows_7:*:sp1:*:*:*:*:*:*

        remove_trailing_regex_stars = r"(?:\:|\:\:|\:\*)+$"

        cpe_regex = re.escape(re.sub(remove_trailing_regex_stars, "", cpe_string))

        cpe_regex_string = r"^{}:".format(cpe_regex)
    else:
        # more general search on same field; e.g. microsoft:windows_7
        cpe_regex_string = f"{re.escape(cpe_string)}"

    return cpe_regex_string



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -----------------------------------------------------------
# CPE format conversions using a WFN object
# Usage example:
# cpe_uri_to_fs("cpe:/a:fmt:fmt:5.2.0") -> "cpe:2.3:a:fmt:fmt:5.2.0:*:*:*:*:*:*:*"
# cpe_fs_to_uri(""cpe:2.3:a:fmt:fmt:5.2.0:*:*:*:*:*:*:*"") -> "cpe:/a:fmt:fmt:5.2.0"
# Reference specification sheet: https://www.govinfo.gov/content/pkg/GOVPUB-C13-c213837a04c3bcc778ebfd420c6a3f2a/pdf/GOVPUB-C13-c213837a04c3bcc778ebfd420c6a3f2a.pdf
# -----------------------------------------------------------

import re


# Convert cpe2.2 url encoded to cpe2.3 char escaped
# cpe:2.3:o:cisco:ios:12.2%281%29 to cpe:2.3:o:cisco:ios:12.2\(1\)
def unquote(cpe: str) -> str:
    return re.compile("%([0-9a-fA-F]{2})", re.M).sub(
        lambda m: "\\" + chr(int(m.group(1), 16)), cpe
    )


# Convert cpe2.3 char escaped to cpe2.2 url encoded
# cpe:2.3:o:cisco:ios:12.2\(1\) to cpe:2.3:o:cisco:ios:12.2%281%29
def quote(cpe: str) -> str:
    cpe = cpe.replace("\\-", "-")
    cpe = cpe.replace("\\.", ".")
    return re.compile("\\\\(.)", re.M).sub(
        lambda m: "%" + hex(ord(m.group(1)))[2:], cpe
    )


class WFN:
    # Default every value to * which is ANY
    part = "*"
    vendor = "*"
    product = "*"
    version = "*"
    update = "*"
    edition = "*"
    language = "*"
    sw_edition = "*"
    target_sw = "*"
    target_hw = "*"
    other = "*"

    # Pack entries unsupported by URI (CPE2.2) representation in to an extended edition string
    # As described in part 6.1.2.1.3 of the specification
    def pack_edition(self) -> str:
        packed_edition = self.edition
        if (
            self.sw_edition != "*"
            or self.target_hw != "*"
            or self.target_sw != "*"
            or self.other != "*"
        ):
            packed_edition = "~{}~{}~{}~{}~{}".format(
                self.edition,
                self.sw_edition,
                self.target_sw,
                self.target_hw,
                self.other,
            ).replace(
                "*", ""
            )  # ANY values in packed edition are empty
        return packed_edition

    # Unpack entries from extended edition string
    def unpack_edition(self, edition_string: str):
        # The packed edition string should always have the same amount of columns if present
        # If there is less its malformed or not a packed edition
        if edition_string.count("~") == 5:
            edition_components = edition_string.split("~")

            if edition_components[1]:
                self.edition = edition_components[1]

            if edition_components[2]:
                self.sw_edition = edition_components[2]

            if edition_components[3]:
                self.target_sw = edition_components[3]

            if edition_components[4]:
                self.target_hw = edition_components[4]

            if edition_components[5]:
                self.other = edition_components[5]
        else:
            self.edition = edition_string

    # Bind values from WFN object to a CPE URI (2.2)
    def bind_to_uri(self) -> str:
        cpe_uri = "cpe:/{}:{}:{}:{}:{}:{}:{}".format(
            self.part,
            self.vendor,
            self.product,
            self.version,
            self.update,
            self.pack_edition(),
            self.language,
        )
        # Any values are * in the object but CPE URIs use empty columns as any
        cpe_uri = cpe_uri.replace(":*", ":").strip(":")
        return cpe_uri

    # Bind values from WFN to a CPE formatted string (2.3)
    def bind_to_fs(self) -> str:
        cpe_fs = "cpe:2.3:{}:{}:{}:{}:{}:{}:{}:{}:{}:{}:{}".format(
            self.part,
            self.vendor,
            self.product,
            self.version,
            self.update,
            self.edition,
            self.language,
            self.sw_edition,
            self.target_sw,
            self.target_hw,
            self.other,
        )
        # Convert sequences like for example "%28" (URI format) to "\(" (FS format)
        cpe_fs = unquote(cpe_fs).strip(":")

        return cpe_fs

    # Unbinds WFN values from a CPE URI (2.2 format)
    def unbind_from_uri(self, cpe_uri: str):
        cpe_components = cpe_uri.split(":")
        # Using a loop to easily deal with variable length CPE uris
        for i, comp in enumerate(cpe_components):
            # URI any values are empty while this object uses * for any
            if comp == "":
                comp = "*"
            if i == 1:
                # Simply splitting by : leaves a stray / in the first column so lets remove it
                self.part = comp.replace("/", "")
            elif i == 2:
                self.vendor = comp
            elif i == 3:
                self.product = comp
            elif i == 4:
                self.version = comp
            elif i == 5:
                self.update = comp
            elif i == 6:
                self.unpack_edition(comp)
            elif i == 7:
                self.language = comp

    # Unbinds WFN values from a CPE FS (2.3 format)
    def unbind_from_fs(self, cpe_fs: str):
        # Convert special character sequences such as "\:" to URI format e.g. "%3a" before splitting
        cpe_fs = quote(cpe_fs)
        cpe_components = cpe_fs.split(":")
        # Using a loop to easily deal with variable length CPE format strings
        for i, comp in enumerate(cpe_components):
            if i == 2:
                self.part = comp
            elif i == 3:
                self.vendor = comp
            elif i == 4:
                self.product = comp
            elif i == 5:
                self.version = comp
            elif i == 6:
                self.update = comp
            elif i == 7:
                self.edition = comp
            elif i == 8:
                self.language = comp
            elif i == 9:
                self.sw_edition = comp
            elif i == 10:
                self.target_sw = comp
            elif i == 11:
                self.target_hw = comp
            elif i == 12:
                self.other = comp


# Convert CPE 2.2 format (URI) to CPE 2.3 format (FS)
def cpe_uri_to_fs(uri: str) -> str:
    wfn = WFN()
    wfn.unbind_from_uri(uri)
    return wfn.bind_to_fs()


# Convert CPE 2.3 format (FS) to CPE 2.2 format (URI)
def cpe_fs_to_uri(fs: str) -> str:
    wfn = WFN()
    wfn.unbind_from_fs(fs)
    return wfn.bind_to_uri()



import click

from CveXplore.cli_cmds.cli_utils.utils import printer
from CveXplore.cli_cmds.mutex_options.mutex import Mutex


@click.group("capec", invoke_without_command=True, help="Query for capec specific data")
@click.pass_context
def capec_cmd(ctx):
    if ctx.invoked_subcommand is None:
        click.echo(capec_cmd.get_help(ctx))


@capec_cmd.group(
    "search",
    invoke_without_command=True,
    help="Search for capec entries by id",
)
@click.option(
    "-c",
    "--capec",
    help="Search for CAPEC's (could be multiple)",
    multiple=True,
)
@click.option(
    "-f",
    "--field",
    help="Field to return (could be multiple)",
    multiple=True,
    cls=Mutex,
    not_required_if=["field_list"],
)
@click.option(
    "-fl",
    "--field_list",
    help="Return a field list for this collection",
    is_flag=True,
    cls=Mutex,
    not_required_if=["field"],
)
@click.option(
    "-o",
    "--output",
    default="json",
    help="Set the desired output format (defaults to json)",
    type=click.Choice(["json", "csv", "xml", "html"], case_sensitive=False),
)
@click.pass_context
def search_cmd(
    ctx,
    capec,
    field,
    field_list,
    output,
):
    if capec and not field_list:
        ret_list = getattr(ctx.obj["data_source"], "capec").mget_by_id(*capec)
    elif capec and field_list:
        ret_list = getattr(ctx.obj["data_source"], "capec").field_list(*capec)
    else:
        click.echo(search_cmd.get_help(ctx))
        return

    if isinstance(ret_list, list) and not field_list:
        result = [result.to_dict(*field) for result in ret_list]
    elif field_list:
        result = [result for result in ret_list]
    else:
        result = []

    if ctx.invoked_subcommand is None:
        printer(input_data=result, output=output)
    else:
        ctx.obj["RESULT"] = result



import datetime
import gzip
import logging
import os
import re
import sys
import tempfile
import threading
import time
import zipfile
from abc import ABC, abstractmethod
from datetime import timedelta
from io import BytesIO
from itertools import islice
from shutil import copy
from typing import Tuple

import requests
from dateutil.parser import parse as parse_datetime
from pymongo.errors import BulkWriteError, InvalidOperation
from requests.adapters import HTTPAdapter
from tqdm.contrib.concurrent import thread_map
from urllib3 import Retry

from CveXplore.common.config import Configuration
from CveXplore.core.general.utils import sanitize
from CveXplore.core.worker_queue.worker_q import WorkerQueue
from ..database_indexer.db_indexer import DatabaseIndexer
from ..logging.handlers.cve_explore_rfh import CveExploreUpdateRfhHandler
from ..logging.handlers.cve_explore_stream import CveExploreUpdateStreamHandler
from ..logging.logger_class import AppLogger
from ...database.connection.database_connection import DatabaseConnection

thread_local = threading.local()
logging.setLoggerClass(AppLogger)

logging.getLogger("urllib3").setLevel(logging.WARNING)


class DownloadHandler(ABC):
    """
    DownloadHandler is the base class for all downloads and subsequent processing of the downloaded content.
    Each download script has a derived class which handles specifics for that type of content / download.
    """

    def __init__(
        self,
        feed_type: str,
        logger_name: str,
        prefix: str = None,
    ):
        self.config = Configuration

        self._end = None

        self.feed_type = feed_type

        self.prefix = prefix

        self.queue = WorkerQueue(name=self.feed_type)

        self.file_queue = WorkerQueue(name=f"{self.feed_type}:files")
        self.file_queue.clear()

        self.progress_bar = None

        self.last_modified = None

        self.do_process = True

        database = DatabaseConnection(
            database_type=self.config.DATASOURCE_TYPE,
            database_init_parameters=self.config.DATASOURCE_CONNECTION_DETAILS,
        ).database_connection

        self.database = database.dbclient

        self.database_indexer = DatabaseIndexer(datasource=database)

        self.logger = logging.getLogger(logger_name)

        if len(self.logger.handlers) == 1:
            self.logger.removeHandler(self.logger.handlers[0])

        self.logger.propagate = False

        self.formatter = logging.Formatter(
            "%(asctime)s - %(name)-8s - %(levelname)-8s - %(message)s"
        )

        crf = None

        cli = CveExploreUpdateStreamHandler(stream=sys.stdout)
        cli.setFormatter(self.formatter)
        cli.setLevel(logging.INFO)

        if self.config.LOGGING_TO_FILE:
            crf = CveExploreUpdateRfhHandler(
                filename=f"{self.config.LOGGING_FILE_PATH}/{self.config.LOGGING_UPDATE_FILE_NAME}",
                maxBytes=self.config.LOGGING_MAX_FILE_SIZE,
                backupCount=self.config.LOGGING_BACKLOG,
            )
            crf.setLevel(logging.DEBUG)
            crf.setFormatter(self.formatter)

        if len(self.logger.handlers) > 0:
            for handler in self.logger.handlers:
                # add the handlers to the logger
                # makes sure no duplicate handlers are added
                if not isinstance(
                    handler, CveExploreUpdateRfhHandler
                ) and not isinstance(handler, CveExploreUpdateStreamHandler):
                    if crf is not None:
                        self.logger.addHandler(crf)
                    self.logger.addHandler(cli)
        else:
            if crf is not None:
                self.logger.addHandler(crf)
            self.logger.addHandler(cli)

    def __repr__(self):
        """return string representation of object"""
        return f"<< {self.__class__.__name__}:{self.feed_type} >>"

    def get_session(
        self,
        retries: int = 5,
        backoff_factor: float = 0.3,
        status_forcelist: tuple = (429, 500, 502, 503, 504),
        session=None,
    ) -> requests.Session:
        """
        Method for returning a session object per every requesting thread
        """

        proxies = self.config.HTTP_PROXY_DICT

        if not hasattr(thread_local, "session"):
            session = session or requests.Session()
            retry = Retry(
                total=retries,
                read=retries,
                connect=retries,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
            )

            session.proxies.update(proxies)

            adapter = HTTPAdapter(max_retries=retry)
            session.mount("http://", adapter)
            session.mount("https://", adapter)

            thread_local.session = session

        return thread_local.session

    def process_downloads(self, sites: list):
        """
        Method to download and process files
        """

        start_time = time.time()

        self.logger.info(
            f"Downloading files (max {self.config.MAX_DOWNLOAD_WORKERS} workers)"
        )

        thread_map(
            self.download_site,
            sites,
            desc="Downloading files",
            max_workers=self.config.MAX_DOWNLOAD_WORKERS,
        )

        if self.do_process:
            thread_map(
                self.file_to_queue,
                self.file_queue.getall(),
                desc="Processing downloaded files",
            )

            chunks = []

            for batch in iter(lambda: list(islice(self.queue, 10000)), []):
                chunks.append(batch)

            thread_map(
                self._db_bulk_writer, chunks, desc="Transferring queue to database"
            )

            # checking if last-modified was in the response headers and not set to default
            if "01-01-1970" != self.last_modified.strftime("%d-%m-%Y"):
                self.setColUpdate(self.feed_type.lower(), self.last_modified)

        self.logger.info(f"Duration: {timedelta(seconds=time.time() - start_time)}")

    def chunk_list(self, lst: list, number: int) -> list:
        """
        Yield successive n-sized chunks from lst.
        """
        for i in range(0, len(lst), number):
            yield lst[i : i + number]

    def clean_keys(self, data):
        """
        Recursively clean dictionary keys for MongoDB by replacing illegal characters with hyphens
        """
        if isinstance(data, dict):
            # Only clean dictionary keys and recurse for the values
            return {
                re.sub(r"[.$\s\+\!\*\&\?]", "-", key): self.clean_keys(value)
                for key, value in data.items()
            }
        elif isinstance(data, list):
            # For lists, just apply clean_keys on items
            return [self.clean_keys(item) for item in data]
        else:
            return data

    def _db_bulk_writer(self, batch: list, initialization_run: bool = False):
        """
        Method to act as worker for writing queued entries into the database
        """

        try:
            if self.feed_type.lower() == "epss":
                self.database["cves"].bulk_write(batch, ordered=False)
            elif initialization_run and (
                self.feed_type.lower() == "cves" or self.feed_type.lower() == "cpe"
            ):
                # cves or cpe process item could yield None; so filter None from batch list
                self.database[self.feed_type.lower()].insert_many(
                    [self.clean_keys(x) for x in batch if x is not None], ordered=False
                )
            else:
                self.database[self.feed_type.lower()].bulk_write(
                    [self.clean_keys(x) for x in batch if x is not None], ordered=False
                )
        except BulkWriteError as err:
            self.logger.debug(f"Error during bulk write: {err}")
            pass
        except InvalidOperation as err:
            self.logger.debug(f"Got error during bulk update: {err}")
            pass
        except TypeError as err:
            self.logger.debug(f"Error during bulk write: {err}")
            raise

    def store_file(
        self, response_content: bytes, content_type: str, url: str
    ) -> Tuple[str, str]:
        """
        Method to store the download based on the headers content type
        """
        wd = tempfile.mkdtemp()
        filename = None
        extension = None

        # If the content type is generic and, therefore, cannot be used for determining
        # the file type, extract the file extension from the url, instead.
        if (
            content_type == "application/octet-stream"
            or content_type == "binary/octet-stream"
            or content_type == "text/plain"
        ):
            extension = url.rsplit(".", 1)[1]
            self.logger.debug(
                f"Content-Type generic '{content_type}'; determining file type based on extension '.{extension}'"
            )

        if (
            content_type == "application/zip"
            or content_type == "application/x-zip"
            or content_type == "application/x-zip-compressed"
            or content_type == "application/zip-compressed"
            or extension == "zip"
        ):
            filename = os.path.join(wd, url.split("/")[-1][:-4])
            self.logger.debug(f"Saving file to: {filename}")

            with zipfile.ZipFile(BytesIO(response_content)) as zip_file:
                zip_file.extractall(wd)

        elif (
            content_type == "application/x-gzip"
            or content_type == "application/gzip"
            or content_type == "application/x-gzip-compressed"
            or content_type == "application/gzip-compressed"
            or extension == "gz"
        ):
            filename = os.path.join(wd, url.split("/")[-1][:-3])
            self.logger.debug(f"Saving file to: {filename}")

            buf = BytesIO(response_content)
            with open(filename, "wb") as f:
                f.write(gzip.GzipFile(fileobj=buf).read())

        elif (
            content_type == "application/json"
            or content_type == "application/xml"
            or content_type == "text/xml"
            or extension == "json"
            or extension == "xml"
        ):
            filename = os.path.join(wd, url.split("/")[-1])
            self.logger.debug(f"Saving file to: {filename}")

            with open(filename, "wb") as output_file:
                output_file.write(response_content)

        elif content_type == "application/local":
            filename = os.path.join(wd, url.split("/")[-1])
            self.logger.debug(f"Saving file to: {filename}")

            copy(url[7:], filename)

        else:
            self.logger.error(
                f"Unhandled Content-Type encountered: {content_type} from url: {url}"
            )
            sys.exit(1)

        return wd, filename

    def source_changed(self, url: str) -> bool:
        """
        Method to check whether the source file has changed since last update using HEAD instead of GET.
        """

        # Always assume a file has changed; only compare last-modified for HTTP resources
        if url[:4] == "file":
            return True
        else:
            i = self.getInfo(self.feed_type.lower())

            if i is None:
                self.logger.info(
                    f"Cached last-modified for {self.feed_type}'s not found; should update"
                )
                return True
            else:
                self.logger.info(
                    f"Comparing cached last-modified of {self.feed_type} ({i['lastModified']}) with URL {url}"
                )
                session = self.get_session()

                try:
                    with session.head(url) as response:
                        try:
                            self.last_modified = parse_datetime(
                                response.headers["last-modified"], ignoretz=True
                            )
                            self.logger.debug(
                                f"Last {self.feed_type} modified value: {self.last_modified} for URL: {url}"
                            )
                        except KeyError:
                            self.logger.info(
                                f"Did not receive last-modified header in the response; assuming {self.feed_type}'s have changed"
                            )
                            return True

                        if self.last_modified == i["lastModified"]:
                            self.logger.info(
                                f"{self.feed_type}'s are not modified since the last update"
                            )
                            return False
                        else:
                            self.logger.info(
                                f"{self.feed_type}'s last-modified changed ({self.last_modified}); should update"
                            )
                            return True

                except Exception:
                    self.logger.warning(
                        f"Exception encountered during comparison; assuming {self.feed_type}'s have changed"
                    )
                    return True

    def download_site(self, url: str):
        if url[:4] == "file":
            self.logger.info(f"Scheduling local hosted file: {url}")

            # local file do not get last_modified header; so completely ignoring last_modified check and always asume
            # local file == the last modified file and set to current time.
            self.last_modified = datetime.datetime.now()

            self.logger.debug(
                f"Last {self.feed_type} modified value: {self.last_modified} for URL: {url}"
            )

            wd, filename = self.store_file(
                response_content=b"local", content_type="application/local", url=url
            )

            if filename is not None:
                self.file_queue.put((wd, filename))
            else:
                self.logger.error(
                    "Unable to retrieve a filename; something went wrong when trying to save the file"
                )
                sys.exit(1)

        else:
            self.logger.debug(f"Downloading from url: {url}")
            session = self.get_session()
            try:
                with session.get(url) as response:
                    try:
                        self.last_modified = parse_datetime(
                            response.headers["last-modified"], ignoretz=True
                        )
                    except KeyError:
                        self.logger.error(
                            f"Did not receive last-modified header in the response; setting to default "
                            f"(01-01-1970) and force update! Headers received: {response.headers}"
                        )
                        # setting to last_modified to default value
                        self.last_modified = parse_datetime("01-01-1970")

                    self.logger.debug(
                        f"Last {self.feed_type} modified value: {self.last_modified} for URL: {url}"
                    )

                    # epss releases on daily basis, if new cve's are inserted in the database
                    # it might be missing epss info, so always run the epss update
                    if self.feed_type.lower() != "epss":
                        i = self.getInfo(self.feed_type.lower())

                        if i is not None:
                            if self.last_modified == i["lastModified"]:
                                self.logger.info(
                                    f"{self.feed_type}'s are not modified since the last update"
                                )
                                self.file_queue.getall()
                                self.do_process = False

                    if self.do_process:
                        content_type = response.headers["content-type"]

                        self.logger.debug(
                            f"URL: {url} fetched Content-Type: {content_type}"
                        )

                        wd, filename = self.store_file(
                            response_content=response.content,
                            content_type=content_type,
                            url=url,
                        )

                        if filename is not None:
                            self.file_queue.put((wd, filename))
                        else:
                            self.logger.error(
                                "Unable to retrieve a filename; something went wrong when trying to save the file"
                            )
                            sys.exit(1)
            except Exception as err:
                self.logger.info(
                    f"Exception encountered during download from: {url}. Please check the logs for more information!"
                )
                self.logger.error(
                    f"Exception encountered during the download from: {url}. Error encountered: {err}"
                )
                self.do_process = False

    def dropCollection(self, col: str):
        return self.database[col].drop()

    def getTableNames(self):
        return self.database.list_collection_names()

    def setColInfo(self, collection: str, field: str, data: dict):
        self.database[collection].update_one(
            {"db": collection}, {"$set": {field: data}}, upsert=True
        )

    def delColInfo(self, collection: str):
        self.database["info"].delete_one({"db": collection})

    def getCPEVersionInformation(self, query: dict):
        return sanitize(self.database["cpe"].find(query))

    def getInfo(self, collection: str):
        return sanitize(self.database["info"].find_one({"db": collection}))

    def setColUpdate(self, collection: str, date: datetime):
        self.database["info"].update_one(
            {"db": collection}, {"$set": {"lastModified": date}}, upsert=True
        )

    @abstractmethod
    def process_item(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def file_to_queue(self, *args):
        raise NotImplementedError

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def populate(self, **kwargs):
        raise NotImplementedError



import click
from tabulate import tabulate

from CveXplore.cli_cmds.mutex_options.mutex import Mutex
from CveXplore.core.general.constants import task_status_rev_types
from CveXplore.core.general.utils import (
    datetimeToTimestring,
    set_ansi_color_red,
    set_ansi_color_green,
    set_ansi_color_magenta,
    set_ansi_color_yellow,
    timestampTOdatetimestring,
)
from CveXplore.errors.tasks import TaskNotFoundError


@click.group(
    "tasks",
    invoke_without_command=True,
    help="Perform task related operations.",
)
@click.pass_context
def tasks_cmd(ctx):
    if ctx.invoked_subcommand is None:
        click.echo(tasks_cmd.get_help(ctx))


@tasks_cmd.group(
    "list",
    invoke_without_command=True,
    help="List the available tasks that can be scheduled.",
)
@click.pass_context
def list_cmd(ctx):
    header_list = ["ID", "Task name", "Task description"]
    x = 1
    table_list = [header_list]
    try:
        for k, v in ctx.obj["data_source"].task_handler.show_available_tasks().items():
            table_list.append([x, k, v])
            x += 1
        click.echo(tabulate(table_list, headers="firstrow", tablefmt="fancy_grid"))
    except Exception as e:
        click.echo(e)


@tasks_cmd.group(
    "scheduled",
    invoke_without_command=True,
    help="Perform action on tasks that are currently scheduled.",
)
@click.option(
    "-l",
    "--list",
    is_flag=True,
    help="List tasks that are scheduled.",
)
@click.option(
    "-d",
    "--delete",
    type=int,
    help="Delete scheduled tasks by task id (id taken from -l / --list command).",
)
@click.option(
    "-t",
    "--toggle",
    type=int,
    help="Toggle scheduled tasks by task id (id taken from -l / --list command) between the enabled and disabled states.",
)
@click.option(
    "-p",
    "--purge",
    type=int,
    help="Purge the scheduled tasks results by task id (id taken from -l / --list command).",
)
@click.option(
    "-r",
    "--results",
    type=int,
    help="Show the (last 10) scheduled task results by task id (id taken from -l / --list command).",
)
@click.pass_context
def scheduled_cmd(ctx, list, delete, toggle, purge, results):

    if list:
        header_list = [
            "ID",
            "Task slug",
            "Task name",
            "Task description",
            "Schedule",
            "Enabled",
            "Last run at [Result] [Total]",
            "Next run at",
        ]
        x = 1
        table_list = [header_list]
        try:
            scheduled_tasks = ctx.obj["data_source"].task_handler.show_scheduled_tasks()
            for each in scheduled_tasks:
                task = each.to_dict()

                try:
                    result = task_status_rev_types[task["last_run_result"]].upper()
                    if result == "OK":
                        last_run_result = f"[{set_ansi_color_green(result)}]"
                    else:
                        last_run_result = f"[{set_ansi_color_red(result)}]"
                except KeyError:
                    last_run_result = "N/A"

                table_list.append(
                    [
                        x,
                        task["name"],
                        task["task"].replace("CveXplore.celery_app.", ""),
                        task["description"],
                        task["run"],
                        (
                            set_ansi_color_green(task["enabled"])
                            if task["enabled"]
                            else set_ansi_color_red(task["enabled"])
                        ),
                        f"{set_ansi_color_magenta(datetimeToTimestring(task['last_run_at']))} "
                        f"{last_run_result} "
                        f"[{set_ansi_color_yellow(task['total_run_count'])}]",
                        set_ansi_color_magenta(
                            datetimeToTimestring(task["next_run_at"])
                        ),
                    ]
                )
                x += 1

            click.echo(tabulate(table_list, headers="firstrow", tablefmt="fancy_grid"))
        except Exception as e:
            click.echo(e)
    elif delete:
        try:
            click.echo(
                ctx.obj["data_source"].task_handler.delete_scheduled_task(delete)
            )
        except TaskNotFoundError:
            click.echo("Task not found; did you specify the correct task number?")
        except Exception as e:
            click.echo(e)
    elif toggle:
        try:
            click.echo(
                ctx.obj["data_source"].task_handler.toggle_scheduled_task(toggle)
            )
        except TaskNotFoundError:
            click.echo("Task not found; did you specify the correct task number?")
        except Exception as e:
            click.echo(e)
    elif purge:
        try:
            click.echo(ctx.obj["data_source"].task_handler.purge_scheduled_task(purge))
        except TaskNotFoundError:
            click.echo("Task not found; did you specify the correct task number?")
        except Exception as e:
            click.echo(e)
    elif results:
        header_list = [
            "ID",
            "Executed at",
            "Task id",
            "Task state",
            "Returned",
            "Task cost",
        ]
        x = 1
        table_list = [header_list]
        try:
            task_results = ctx.obj[
                "data_source"
            ].task_handler.get_scheduled_tasks_results(results)
            for each in task_results:
                table_list.append(
                    [
                        x,
                        set_ansi_color_magenta(
                            timestampTOdatetimestring(each["inserted"])
                        ),
                        each["task_id"],
                        each["state"],
                        each["returns"],
                        set_ansi_color_yellow(each["cost"]),
                    ]
                )
                x += 1

            click.echo(tabulate(table_list, headers="firstrow", tablefmt="fancy_grid"))
        except TaskNotFoundError:
            click.echo("Task not found; did you specify the correct task number?")
        except Exception as e:
            click.echo(e)
    else:
        if ctx.invoked_subcommand is None:
            click.echo(tasks_cmd.get_help(ctx))


@tasks_cmd.group(
    "create",
    invoke_without_command=True,
    help="Perform action on tasks that are currently scheduled.",
)
@click.option(
    "-n",
    "--number",
    type=int,
    required=True,
    help="Create new task of type referred to by task id (id taken from 'cvexplore tasks list' command).",
)
@click.option(
    "-s",
    "--slug",
    required=True,
    help="Slug of the task to create.",
)
@click.option(
    "-i",
    "--interval",
    help="Use interval as tasks schedule; interval is in seconds.",
    type=int,
    cls=Mutex,
    not_required_if=["crontab"],
)
@click.option(
    "-c",
    "--crontab",
    help="Use crontab as tasks schedule; use csv value in format 'minute,hour,day_of_week,day_of_month,month_of_year'"
    "The crontab entry will be processed from left to right (from minute to month_of_year) with a '*' as a "
    "default entry. That means that is you only need to set the minute value to every minute -c */1 will suffice.",
    cls=Mutex,
    not_required_if=["interval"],
)
@click.pass_context
def create_cmd(ctx, number, slug, interval, crontab):
    if interval:
        try:
            click.echo(
                ctx.obj["data_source"].task_handler.create_task_by_number(
                    task_number=number, task_slug=slug, task_interval=interval
                )
            )
        except TaskNotFoundError:
            click.echo("Task not found; did you specify the correct task number?")
        except Exception as e:
            click.echo(e)
    elif crontab:

        the_crontab = {
            "minute": "*",
            "hour": "*",
            "day_of_week": "*",
            "day_of_month": "*",
            "month_of_year": "*",
        }

        crontab_list = crontab.split(",")
        try:
            i = 0
            for each in crontab_list:
                the_crontab[list(the_crontab.keys())[i]] = each
                i += 1

            click.echo(
                ctx.obj["data_source"].task_handler.create_task_by_number(
                    task_number=number, task_slug=slug, task_crontab=the_crontab
                )
            )
        except Exception as err:
            click.echo(
                f"Could not insert crontab; error: {err}",
            )
    else:
        if ctx.invoked_subcommand is None:
            click.echo(tasks_cmd.get_help(ctx))



from CveXplore.objects.capec import Capec
from CveXplore.objects.cpe import Cpe
from CveXplore.objects.cves import Cves
from CveXplore.objects.cwe import Cwe
from CveXplore.objects.via4 import Via4

database_objects_mapping = {
    "capec": Capec,
    "cpe": Cpe,
    "cwe": Cwe,
    "via4": Via4,
    "cves": Cves,
}



import click

from CveXplore.cli_cmds.cli_utils.utils import printer


@click.group(
    "stats",
    invoke_without_command=True,
    help="Show datasource statistics",
)
@click.pass_context
def stats_cmd(ctx, output="json"):
    if ctx.invoked_subcommand is None:
        printer(
            input_data=[ctx.obj["data_source"].get_db_content_stats()],
            output=output,
        )



from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase


class DummyConnection(DatabaseConnectionBase):
    def __init__(self, **kwargs):
        super().__init__(logger_name=__name__)

        self._dbclient = {"schema": "test"}

    @property
    def dbclient(self):
        return self._dbclient

    def set_handlers_for_collections(self):
        pass



from CveXplore.common.data_source_connection import DatasourceConnection


class Via4(DatasourceConnection):
    """
    Via4 database object
    """

    def __init__(self, **kwargs):
        super().__init__("via4")

        for each in kwargs:
            setattr(self, each, kwargs[each])

    def __repr__(self):
        """String representation of object"""
        return f"<< Via4:{self.id} >>"



from CveXplore.common.config import Configuration


class CveXploreObject(object):
    """
    CveXploreObject is the base object for all database collection objects
    """

    def __init__(self):
        self.config = Configuration

    def __repr__(self) -> str:
        return f"<< {self.__class__.__name__} >>"



import logging
import os
from logging.handlers import RotatingFileHandler

import colors

from CveXplore.common.config import Configuration
from CveXplore.core.logging.formatters.task_formatter import TaskFormatter
from CveXplore.core.logging.handlers.gelf_handler import GelfUDPHandler

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

level_map = {
    "debug": "magenta",
    "info": "white",
    "warning": "yellow",
    "error": "red",
    "critical": "red",
}


class AppLogger(logging.Logger):
    def __init__(self, name, level=logging.NOTSET):
        self.formatter = TaskFormatter(
            "%(asctime)s - %(task_name)s - %(name)-8s - %(levelname)-8s - [%(task_id)s] %(message)s"
        )
        self.config = Configuration

        root = logging.getLogger()

        root.setLevel(logging.DEBUG)

        root_null_handler = logging.NullHandler()
        root.handlers.clear()
        root.addHandler(root_null_handler)

        super().__init__(name, level)

        self.propagate = False

        if self.config.LOGGING_FILE_PATH != "":
            if not os.path.exists(self.config.LOGGING_FILE_PATH):
                os.makedirs(self.config.LOGGING_FILE_PATH)

            crf = RotatingFileHandler(
                filename=f"{self.config.LOGGING_FILE_PATH}/{self.config.LOGGING_FILE_NAME}",
                maxBytes=self.config.LOGGING_MAX_FILE_SIZE,
                backupCount=self.config.LOGGING_BACKLOG,
            )
            crf.setLevel(self.config.LOGGING_LEVEL)
            crf.setFormatter(self.formatter)
            self.addHandler(crf)

        if self.config.SYSLOG_ENABLE:
            syslog_server = self.config.SYSLOG_SERVER
            syslog_port = self.config.SYSLOG_PORT

            if self.config.GELF_SYSLOG:
                syslog = GelfUDPHandler(
                    host=syslog_server,
                    port=syslog_port,
                    _application_name="CveXplore",
                    include_extra_fields=True,
                    debug=True,
                    **(
                        self.config.GELF_SYSLOG_ADDITIONAL_FIELDS
                        if self.config.GELF_SYSLOG_ADDITIONAL_FIELDS is not None
                        else {}
                    ),
                )
            else:
                # Import syslog_handler only if configured (for Windows compatibility)
                from CveXplore.core.logging.handlers.syslog_handler import (
                    FullSysLogHandler,
                )

                syslog = FullSysLogHandler(
                    address=(syslog_server, syslog_port),
                    facility=FullSysLogHandler.LOG_LOCAL0,
                    appname="CveXplore",
                )

            syslog.setFormatter(self.formatter)
            syslog.setLevel(self.config.SYSLOG_LEVEL)
            self.addHandler(syslog)

    def debug(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘DEBUG’ and color *MAGENTA.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.debug(“Houston, we have a %s”, “thorny problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=level_map["debug"])

        if self.isEnabledFor(DEBUG):
            self._log(DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘INFO’ and color *WHITE*.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.info(“Houston, we have a %s”, “interesting problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=level_map["info"])

        if self.isEnabledFor(INFO):
            self._log(INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘WARNING’ and color *YELLOW*.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.warning(“Houston, we have a %s”, “bit of a problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=level_map["warning"])

        if self.isEnabledFor(WARNING):
            self._log(WARNING, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘ERROR’ and color *RED*.

        Store logged message to the database for dashboard alerting.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.error(“Houston, we have a %s”, “major problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=level_map["error"])

        if self.isEnabledFor(ERROR):
            self._log(ERROR, msg, args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log ‘msg % args’ with severity ‘CRITICAL’ and color *RED*.

        Store logged message to the database for dashboard alerting.

        To pass exception information, use the keyword argument exc_info with a true value, e.g.

        logger.critical(“Houston, we have a %s”, “hell of a problem”, exc_info=1)

        :param msg: Message to log
        :type msg: str
        """

        msg = colors.color("{}".format(msg), fg=level_map["critical"])

        if self.isEnabledFor(CRITICAL):
            self._log(CRITICAL, msg, args, **kwargs)



import atexit

from pymongo import MongoClient
from pymongo.errors import ServerSelectionTimeoutError

from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase
from CveXplore.database.helpers.cvesearch_mongo_database import CveSearchCollection
from CveXplore.errors import DatabaseConnectionException


class MongoDBConnection(DatabaseConnectionBase):
    """
    The MongoDBConnection class serves as a shell that functions as uniform way to connect to the mongodb backend.
    By default, it will try to establish a connection towards a mongodb running on localhost (default port 27017) and
    database 'cvedb' (as per defaults of cve_search)

    Group:
        database
    """

    def __init__(
        self,
        host: str = "mongodb://127.0.0.1:27017",
        port: int = None,
        database: str = "cvedb",
        **kwargs,
    ):
        """
        The `host` parameter can be a full `mongodb URI <http://dochub.mongodb.org/core/connections>`_, in addition to
        a simple hostname.
        """
        super().__init__(logger_name=__name__)

        self.client = None
        self._dbclient = None

        self.client = MongoClient(host, port, connect=False, **kwargs)

        self._dbclient = self.client[database]

        try:
            collections = self.dbclient.list_collection_names()
        except ServerSelectionTimeoutError as err:
            raise DatabaseConnectionException(
                f"Connection to the database failed: {err}"
            )

        if len(collections) != 0:
            for each in collections:
                self.__setattr__(
                    f"store_{each}",
                    CveSearchCollection(database=self.dbclient, name=each),
                )

        atexit.register(self.disconnect)

    @property
    def dbclient(self):
        return self._dbclient

    def set_handlers_for_collections(self):
        for each in self.dbclient.list_collection_names():
            if not hasattr(self, each):
                setattr(
                    self,
                    f"store_{each}",
                    CveSearchCollection(database=self.dbclient, name=each),
                )

    def disconnect(self):
        """
        Disconnect from mongodb
        """
        if self.client is not None:
            self.client.close()

    def __del__(self):
        """Called when the class is garbage collected."""
        self.disconnect()



import json
import os

from CveXplore.core.database_maintenance.update_base_class import UpdateBaseClass
from CveXplore.database.connection.base.db_connection_base import DatabaseConnectionBase
from CveXplore.errors import DatabaseSchemaVersionError

runPath = os.path.dirname(os.path.realpath(__file__))


class DatabaseVersionChecker(UpdateBaseClass):
    def __init__(self, datasource: DatabaseConnectionBase):
        super().__init__(logger_name=__name__)

        with open(os.path.join(runPath, "../../.schema_version")) as f:
            self.schema_version = json.loads(f.read())

        database = datasource

        self.dbh = database.dbclient["schema"]

    def validate_schema(self):
        try:
            if (
                not self.schema_version["version"]
                == list(self.dbh.find({}))[0]["version"]
            ):
                if not self.schema_version["rebuild_needed"]:
                    raise DatabaseSchemaVersionError(
                        "Database is not on the latest schema version; please update the database!"
                    )
                else:
                    raise DatabaseSchemaVersionError(
                        "Database schema is not up to date; please re-populate the database!"
                    )
            else:
                return True
        except IndexError:
            # something went wrong fetching the result from the database; assume re-populate is needed
            raise DatabaseSchemaVersionError(
                "Database schema is not up to date; please re-populate the database!"
            )

    def update(self):
        self.logger.info("Updating schema version")

        try:
            current_record = list(self.dbh.find({}))

            if len(current_record) != 0:
                current_record[0]["version"] = self.schema_version["version"]
                current_record[0]["rebuild_needed"] = self.schema_version[
                    "rebuild_needed"
                ]

                self.dbh.update_one(
                    {"_id": current_record[0]["_id"]}, {"$set": current_record[0]}
                )
            else:
                current_record = {
                    "version": self.schema_version["version"],
                    "rebuild_needed": self.schema_version["rebuild_needed"],
                }

                self.dbh.insert_one(current_record)
        except AttributeError:
            current_record = {
                "version": self.schema_version["version"],
                "rebuild_needed": self.schema_version["rebuild_needed"],
            }
            self.dbh.insert_one(current_record)

        self.logger.info("Update schema version done!")



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from CveXplore.common.config import Configuration

config = Configuration

engine = create_engine(config.SQLALCHEMY_DATABASE_URI, echo=True)

Session = sessionmaker(bind=engine)



import logging
import uuid
from contextlib import contextmanager

from redis import Redis

from CveXplore.common.config import Configuration


class RedisHandler(object):
    """
    Redis Handler for CveXplore which handlers several Redis related operations.

    Group:
        backend
    """

    def __init__(self, redis_client: Redis):
        """
        Create a new RedisHandler instance

        Args:
            redis_client: A Redis client instance
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.config = Configuration

        self.redis_client = redis_client

    def _request_lock(self, lock_key: str) -> bool:
        """
        Method to request a lock. A lock is set into Redis and takes the `lock_key` as an argument for it's key.
        Lock is set to the `CELERY_RESULT_EXPIRES` configuration variable.

        Args:
            lock_key: The string to use as the lock key.

        Returns:
            True if lock is acquired, False otherwise
        Group:
            backend
        """
        got_lock = False
        while not got_lock:
            if self.redis_client.get(lock_key) is None:
                self.redis_client.set(
                    lock_key,
                    f"{uuid.uuid4()}",
                    ex=self.config.CELERY_RESULT_EXPIRES,
                )
                got_lock = True
            else:
                got_lock = False

        self.logger.info(f"Lock request: {lock_key} -> Lock request result: {got_lock}")

        return got_lock

    @contextmanager
    def acquire_lock(self, lock_key: str) -> bool:
        """
        Context manager to acquire a lock; lock is deleted when context manager exits.

        Args:
            lock_key:

        Returns:
            True if lock is acquired, False otherwise
        """
        lock_acquired = self._request_lock(lock_key=lock_key)

        try:
            self.logger.info(f"Lock acquired! -> {lock_key}")
            yield lock_acquired
        finally:
            if lock_acquired:
                self.redis_client.delete(lock_key)
                self.logger.info(f"Releasing lock: {lock_key}")



from CveXplore.api.helpers.cve_search_api import CveSearchApi
from CveXplore.common.db_mapping import database_mapping


class ApiDatabaseSource(object):
    """
    The ApiDatabaseSource mimics the behaviour of the MongoDBConnection.
    """

    def __init__(
        self,
        baseurl: str,
        api_path: str = None,
        proxies: dict = None,
        user_agent: str = "CveXplore",
    ):
        """
        Create new instance of the ApiDatabaseSource
        """

        self.database_mapping = database_mapping

        self.baseurl = baseurl

        for each in self.database_mapping:
            setattr(
                self,
                f"store_{each}",
                ApiDatabaseCollection(
                    baseurl=baseurl,
                    api_path=api_path,
                    proxies=proxies,
                    user_agent=user_agent,
                    collname=each,
                ),
            )

    def __repr__(self):
        """return a string representation of the obj ApiDatabaseSource"""
        return f"<< ApiDatabaseSource: {self.baseurl} >>"


class ApiDatabaseCollection(object):
    """
    The ApiDatabaseCollection mimics the behaviour of the CveSearchCollection
    """

    def __init__(
        self,
        baseurl: str,
        collname: str,
        api_path: str = None,
        proxies: dict = None,
        user_agent="CveXplore",
    ):
        """
        Create a new ApiDatabaseCollection.
        """

        self.baseurl = baseurl
        self.api_path = api_path
        self.proxies = proxies
        self.user_agent = user_agent

        self.collname = collname

    def find(self, the_filter: dict = None):
        """
        Query the api endpoint as you would do so with a pymongo Collection.
        """

        return CveSearchApi(self, the_filter)

    def find_one(self, the_filter: dict = None):
        """
        Query the api endpoint as you would do so with a pymongo Collection.

        :return: Data or None
        :rtype: object
        """

        cursor = self.find(the_filter)

        for result in cursor.limit(-1):
            return result
        return None

    def __repr__(self):
        """return a string representation of the obj ApiDatabaseCollection"""
        return f"<< ApiDatabaseCollection: {self.baseurl} >>"
